/*
  Copyright (C) 2001 Kimmo Pekkola

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
/*
  $Header: //rainbox/cvsroot/Rainmeter/Plugin/MeterWindow.h,v 1.17 2003/02/10 18:10:56 rainy Exp $

  $Log: MeterWindow.h,v $
  Revision 1.17  2003/02/10 18:10:56  rainy
  no message

  Revision 1.16  2002/12/23 14:25:21  rainy
  Separated skin and other settings.
  Stats are now written in ini file.

  Revision 1.15  2002/07/01 15:27:36  rainy
  AlwaysOnTop gets now three values.
  Added Toggle().
  Added RainmeterCurrentConfig and RainmeterCurrentConfigIni.
  The ini file's timestamp is checked before it is overwritten.
  AboutBox is now in separate source file.
  Added SnapEdges.

  Revision 1.14  2002/05/05 10:48:15  rainy
  The last message is stored into a membervariable.

  Revision 1.13  2002/05/04 08:14:45  rainy
  Configs store also the names of the ini-files.

  Revision 1.12  2002/04/27 10:27:42  rainy
  Added hide/show to meters and enable/disable to measures

  Revision 1.11  2002/03/31 09:58:53  rainy
  Added some comments

  Revision 1.10  2001/12/23 10:14:09  rainy
  Added support for different configs.
  The position of the window is now remembered.

  Revision 1.9  2001/10/28 10:15:21  rainy
  Added left and right mouse up actions.
  IsNT() can now identify different OS more precisely.

  Revision 1.8  2001/10/14 07:26:21  rainy
  Added m_StatsDate

  Revision 1.7  2001/09/26 16:25:44  rainy
  Added support for statistics.
  Meters and Measures are now stored here.

  Revision 1.6  2001/09/01 12:57:13  rainy
  Added support for Bar and Bitmap meters.

  Revision 1.5  2001/08/25 18:08:34  rainy
  Added mousebutton actions.
  The About dialog has now the build date.

  Revision 1.4  2001/08/25 17:14:00  rainy
  Pointer to CRainmeter is now stored in a member variable.
  Added few more methods.

  Revision 1.3  2001/08/19 09:10:18  rainy
  Added GetWindow() and OnGetRevID().

  Revision 1.2  2001/08/12 15:37:39  Rainy
  ToggleWindowIfNecessary is now ShowWindowIfAppropriate.
  Added method for mouseover messages.

  Revision 1.1.1.1  2001/08/11 10:58:19  Rainy
  Added to CVS.

*/

#ifndef __METERWINDOW_H__
#define __METERWINDOW_H__

#pragma warning(disable: 4786)

#include <windows.h>
#include <gdiplus.h>
#include <string>
#include <list>
#include <vector>

enum PLATFORM
{
	PLATFORM_9X,
	PLATFORM_NT4,
	PLATFORM_2K,
	PLATFORM_XP
};

enum MOUSE
{
	MOUSE_LMB_DOWN,
	MOUSE_LMB_UP,
	MOUSE_RMB_DOWN,
	MOUSE_RMB_UP
};

enum ZPOSITION
{
	ZPOSITION_ONBOTTOM = -1,
	ZPOSITION_NORMAL = 0,
	ZPOSITION_ONTOP = 1
};

enum BGMODE
{
	BGMODE_IMAGE = 0,
	BGMODE_COPY,
	BGMODE_SOLID
};

enum BEVELTYPE 
{
	BEVELTYPE_NONE,
	BEVELTYPE_UP,
	BEVELTYPE_DOWN
};

struct GlobalConfig
{
	UINT netInSpeed;
	UINT netOutSpeed;
};

class CRainmeter;
class CMeasure;
class CMeter;

class CMeterWindow
{
public:
	CMeterWindow();
	~CMeterWindow();

	int Initialize(CRainmeter& Rainmeter, HWND Parent, HINSTANCE Instance);

	void Hide();
	void Show();
	void Toggle();
	void HideMeter(const char* name);
	void ShowMeter(const char* name);
	void ToggleMeter(const char* name);
	void DisableMeasure(const char* name);
	void EnableMeasure(const char* name);
	void ToggleMeasure(const char* name);
	void Refresh(bool init, const std::string configName = "", const std::string configIniFile = "");
	void Redraw();

	void MoveWindow(int x, int y);
	void ChangeZPos(ZPOSITION zPos);
	
	Gdiplus::Bitmap* GetDoubleBuffer() { return m_DoubleBuffer; };
	HWND GetWindow() { return m_Window; };
	HINSTANCE GetInstance() { return m_Instance; };

	PLATFORM IsNT();
	void ExecuteCommand(const char* command);

	std::list<CMeasure*>& GetMeasures() { return m_Measures; };
	std::list<CMeter*>& GetMeters() { return m_Meters; };

	const std::string& GetStatsDate() { return m_StatsDate; };

	static const GlobalConfig& GetGlobalConfig() { return c_GlobalConfig; };

protected:
	static LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

	LRESULT OnPaint(WPARAM wParam, LPARAM lParam);
	LRESULT OnMove(WPARAM wParam, LPARAM lParam);
	LRESULT OnCreate(WPARAM wParam, LPARAM lParam);
	LRESULT OnDestroy(WPARAM wParam, LPARAM lParam);
	LRESULT OnTimer(WPARAM wParam, LPARAM lParam);
	LRESULT OnCommand(WPARAM wParam, LPARAM lParam);
	LRESULT OnNcHitTest(WPARAM wParam, LPARAM lParam);
	LRESULT OnWindowPosChanging(WPARAM wParam, LPARAM lParam);
	LRESULT OnMouseMove(WPARAM wParam, LPARAM lParam);
	LRESULT OnContextMenu(WPARAM wParam, LPARAM lParam);
	LRESULT OnGetRevID(WPARAM wParam, LPARAM lParam);
	LRESULT OnLeftButtonDown(WPARAM wParam, LPARAM lParam);
	LRESULT OnRightButtonDown(WPARAM wParam, LPARAM lParam);
	LRESULT OnLeftButtonUp(WPARAM wParam, LPARAM lParam);
	LRESULT OnRightButtonUp(WPARAM wParam, LPARAM lParam);
	LRESULT OnCopyData(WPARAM wParam, LPARAM lParam);

private:
	void CreateRegion(bool clear);
	void GetSkinFolders(const std::string& folder);
	Gdiplus::Bitmap* GrabDesktop(int x, int y, int w, int h);

	void ReadStats();
	void WriteStats();

	void QuitRainmeter();

	void Update();
	void UpdateTransparency(bool reset);
	void GetCurrentConfig(LPCSTR cmdLine);
	void SelectConfig(const std::string& configName, const std::string& configIniFile);
	void ReadConfig(const std::string& iniFile, bool generalConfig);
	void ReadSkin(const std::string& iniFile);
	void WriteConfig(const std::string& iniFile);
	void InitializeMeters();
	void ShowWindowIfAppropriate();
	bool DoAction(int x, int y, MOUSE mouse, bool test);

	Gdiplus::Bitmap* m_DoubleBuffer;						// Double buffer for flicker free drawing
	Gdiplus::Bitmap* m_Background;						// The background bitmap

	HWND m_Window;								// Handle to the Rainmeter window
	HINSTANCE m_Instance;						// Handle to the DLL
	HINSTANCE m_User32Library;

	std::string m_RightMouseDownAction;			// Action to run when right mouse is pressed
	std::string m_LeftMouseDownAction;			// Action to run when left mouse is pressed
	std::string m_RightMouseUpAction;			// Action to run when right mouse is released
	std::string m_LeftMouseUpAction;			// Action to run when left mouse is released

	std::string m_BackgroundName;				// Name of the background image
	int m_WindowX;								// Window's X-position
	int m_WindowY;								// Window's Y-position
	int m_WindowW;								// Window's width
	int m_WindowH;								// Window's Height
	bool m_WindowDraggable;						// True, if window can be moved
	int m_WindowUpdate;							// Measure update frequency
	bool m_WindowHide;							// If true, the window is hidden when mouse is over it
	bool m_WindowStartHidden;					// If true, the window is hidden at startup
	bool m_SavePosition;						// If true, the window's position is saved
	bool m_SnapEdges;							// If true, the window snaps to the edges of the screen when moved
	bool m_NativeTransparency;					// If true, use the W2k/XP native transparency
	bool m_MeasuresToVariables;					// If true, Measured values are transformed to Litestep's eVars
	bool m_AllowNegativeCoordinates;			// If true, Negative coordinates are OK
	ZPOSITION m_WindowZPosition;				// Window's Z-position

	BGMODE m_BackgroundMode;					// The background mode
	Gdiplus::Color m_SolidColor;				// Color of the solid background
	BEVELTYPE m_SolidBevel;						// The type of the bevel

	bool m_Refreshing;							// This is true, when the meter is refreshing

	bool m_Hidden;								// True, if Rainmeter is hidden
	bool m_ResetRegion;							// If true, the window region is recalculated during the next update

	std::string m_StatsDate;					// The date when stats gathering started

	std::list<CMeasure*> m_Measures;			// All the measures
	std::list<CMeter*> m_Meters;				// All the meters

	std::string m_CurrentConfig;				// Filename and path of the the current config
	std::string m_SkinName;						// Name of the current skin folder
	std::string m_SkinIniFile;					// Name of the current skin iniFile
	std::string m_SkinPath;						// Path to the folder where the skins are

	UINT m_Message;								// The current window message

	struct CONFIG 
	{
		std::string path;
		std::vector<std::string> iniFiles;
	};

	std::vector<CONFIG> m_ConfigStrings;	    // All configs found in the given folder
	static GlobalConfig c_GlobalConfig;

	CRainmeter* m_Rainmeter;					// Pointer to the main object

	ULONG_PTR           m_GDIplusToken;
};

#endif
